import os
import sqlite3
from contextlib import closing
from typing import Dict, List, Optional, Tuple

from specklepy.core.helpers import speckle_path_provider
from specklepy.logging.exceptions import SpeckleException
from specklepy.transports.abstract_transport import AbstractTransport


class SQLiteTransport(AbstractTransport):
    def __init__(
        self,
        base_path: Optional[str] = None,
        app_name: Optional[str] = None,
        scope: Optional[str] = None,
        max_batch_size_mb: float = 10.0,
        name: str = "SQLite",
    ) -> None:
        super().__init__()
        self._name = name
        self.app_name = app_name or "Speckle"
        self.scope = scope or "Objects"
        self._base_path = base_path or self.get_base_path(self.app_name)
        self.max_size = int(max_batch_size_mb * 1000 * 1000)
        self.saved_obj_count = 0
        self._current_batch: List[Tuple[str, str]] = []
        self._current_batch_size = 0

        try:
            os.makedirs(self._base_path, exist_ok=True)

            self._root_path = os.path.join(
                os.path.join(self._base_path, f"{self.scope}.db")
            )
            self.__initialise()
        except Exception as ex:
            raise SpeckleException(
                f"SQLiteTransport could not initialise {self.scope}.db at"
                f" {self._base_path}. Either provide a different `base_path` or use an"
                " alternative transport.",
                ex,
            )

    def __repr__(self) -> str:
        return f"SQLiteTransport(app: '{self.app_name}', scope: '{self.scope}')"

    @property
    def name(self) -> str:
        return self._name

    @staticmethod
    def get_base_path(app_name):
        return str(
            speckle_path_provider.user_application_data_path().joinpath(app_name)
        )

    def save_object_from_transport(
        self, id: str, source_transport: AbstractTransport
    ) -> None:
        """Adds an object from the given transport to the the local db

        Arguments:
            id {str} -- the object id
            source_transport {AbstractTransport)
            -- the transport through which the object can be found
        """
        serialized_object = source_transport.get_object(id)
        self.save_object(id, serialized_object)

    def save_object(self, id: str, serialized_object: str) -> None:
        """
        Adds an object to the current batch to be written to the db.
        If the current batch is full,
        the batch is written to the db and the current batch is reset.

        Arguments:
            id {str} -- the object id
            serialized_object {str} -- the full string representation of the object
        """
        obj_size = len(serialized_object)
        if (
            not self._current_batch
            or self._current_batch_size + obj_size < self.max_size
        ):
            self._current_batch.append((id, serialized_object))
            self._current_batch_size += obj_size
            return

        self.save_current_batch()
        self._current_batch = [(id, serialized_object)]
        self._current_batch_size = obj_size

    def save_current_batch(self) -> None:
        """Save the current batch of objects to the local db"""
        self.__check_connection()
        try:
            with closing(self.__connection.cursor()) as c:
                c.executemany(
                    "INSERT OR IGNORE INTO objects(hash, content) VALUES(?,?)",
                    self._current_batch,
                )
                self.__connection.commit()
        except Exception as ex:
            raise SpeckleException(
                "Could not save the batch of objects to the local db. Inner exception:"
                f" {ex}",
                ex,
            )

    def get_object(self, id: str) -> str or None:
        self.__check_connection()
        with closing(self.__connection.cursor()) as c:
            row = c.execute(
                "SELECT * FROM objects WHERE hash = ? LIMIT 1", (id,)
            ).fetchone()
        return row[1] if row else None

    def has_objects(self, id_list: List[str]) -> Dict[str, bool]:
        ret = {}
        self.__check_connection()
        with closing(self.__connection.cursor()) as c:
            for id in id_list:
                row = c.execute(
                    "SELECT 1 FROM objects WHERE hash = ? LIMIT 1", (id,)
                ).fetchone()
                ret[id] = bool(row)
        return ret

    def begin_write(self):
        self._object_cache = []
        self.saved_obj_count = 0

    def end_write(self):
        if self._current_batch:
            self.save_current_batch()
        self._current_batch = []
        self._current_batch_size = 0

    def copy_object_and_children(
        self, id: str, target_transport: AbstractTransport
    ) -> str:
        raise NotImplementedError

    def get_all_objects(self):
        """
        Returns all the objects in the store.
        NOTE: do not use for large collections!
        """
        self.__check_connection()
        with closing(self.__connection.cursor()) as c:
            rows = c.execute("SELECT * FROM objects").fetchall()
        return rows

    def close(self):
        """Close the connection to the database"""
        if self.__connection:
            self.__connection.close()
            self.__connection = None

    def __initialise(self) -> None:
        self.__connection = sqlite3.connect(self._root_path)
        with closing(self.__connection.cursor()) as c:
            c.execute(
                """ CREATE TABLE IF NOT EXISTS objects(
                      hash TEXT PRIMARY KEY,
                      content TEXT
                    ) WITHOUT ROWID;"""
            )
            c.execute("PRAGMA journal_mode='wal';")
            c.execute("PRAGMA count_changes=OFF;")
            c.execute("PRAGMA temp_store=MEMORY;")
            self.__connection.commit()

    def __check_connection(self):
        if not self.__connection:
            self.__connection = sqlite3.connect(self._root_path)

    def __del__(self):
        self.close()



from typing import Optional, Union

from deprecated import deprecated

from specklepy.api.models import Branch
from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
)
from specklepy.core.api.resources.deprecated.branch import Resource as CoreResource
from specklepy.logging import metrics
from specklepy.logging.exceptions import SpeckleException


class Resource(CoreResource):
    """API Access class for branches"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
        )
        self.schema = Branch

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def create(
        self, stream_id: str, name: str, description: str = "No description provided"
    ) -> str:
        """Create a new branch on this stream

        Arguments:
            name {str} -- the name of the new branch
            description {str} -- a short description of the branch

        Returns:
            id {str} -- the newly created branch's id
        """
        metrics.track(metrics.SDK, self.account, {"name": "Branch Create"})
        return super().create(stream_id, name, description)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get(
        self, stream_id: str, name: str, commits_limit: int = 10
    ) -> Union[Branch, None, SpeckleException]:
        """Get a branch by name from a stream

        Arguments:
            stream_id {str} -- the id of the stream to get the branch from
            name {str} -- the name of the branch to get
            commits_limit {int} -- maximum number of commits to get

        Returns:
            Branch -- the fetched branch with its latest commits
        """
        metrics.track(metrics.SDK, self.account, {"name": "Branch Get"})
        return super().get(stream_id, name, commits_limit)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def list(self, stream_id: str, branches_limit: int = 10, commits_limit: int = 10):
        """Get a list of branches from a given stream

        Arguments:
            stream_id {str} -- the id of the stream to get the branches from
            branches_limit {int} -- maximum number of branches to get
            commits_limit {int} -- maximum number of commits to get

        Returns:
            List[Branch] -- the branches on the stream
        """
        metrics.track(metrics.SDK, self.account, {"name": "Branch List"})
        return super().list(stream_id, branches_limit, commits_limit)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def update(
        self,
        stream_id: str,
        branch_id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """Update a branch

        Arguments:
            stream_id {str} -- the id of the stream containing the branch to update
            branch_id {str} -- the id of the branch to update
            name {str} -- optional: the updated branch name
            description {str} -- optional: the updated branch description

        Returns:
            bool -- True if update is successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Branch Update"})
        return super().update(stream_id, branch_id, name, description)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def delete(self, stream_id: str, branch_id: str):
        """Delete a branch

        Arguments:
            stream_id {str} -- the id of the stream containing the branch to delete
            branch_id {str} -- the branch to delete

        Returns:
            bool -- True if deletion is successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Branch Delete"})
        return super().delete(stream_id, branch_id)



from typing import Any, Dict, List, Tuple

from specklepy.api.models import ServerInfo
from specklepy.core.api.resources import ServerResource as CoreResource
from specklepy.logging import metrics


class ServerResource(CoreResource):
    """API Access class for the server"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
        )

    def get(self) -> ServerInfo:
        """Get the server info

        Returns:
            dict -- the server info in dictionary form
        """
        metrics.track(metrics.SDK, self.account, {"name": "Server Get"})
        return super().get()

    def version(self) -> Tuple[Any, ...]:
        """Get the server version

        Returns:
            the server version in the format (major, minor, patch, (tag, build))
            eg (2, 6, 3) for a stable build and (2, 6, 4, 'alpha', 4711) for alpha
        """
        # not tracking as it will be called along with other mutations / queries as a check
        return super().version()

    def apps(self) -> Dict:
        """Get the apps registered on the server

        Returns:
            dict -- a dictionary of apps registered on the server
        """
        metrics.track(metrics.SDK, self.account, {"name": "Server Apps"})
        return super().apps()

    def create_token(self, name: str, scopes: List[str], lifespan: int) -> str:
        """Create a personal API token

        Arguments:
            scopes {List[str]} -- the scopes to grant with this token
            name {str} -- a name for your new token
            lifespan {int} -- duration before the token expires

        Returns:
            str -- the new API token. note: this is the only time you'll see the token!
        """
        metrics.track(metrics.SDK, self.account, {"name": "Server Create Token"})
        return super().create_token(name, scopes, lifespan)

    def revoke_token(self, token: str) -> bool:
        """Revokes (deletes) a personal API token

        Arguments:
            token {str} -- the token to revoke (delete)

        Returns:
            bool -- True if the token was successfully deleted
        """
        metrics.track(metrics.SDK, self.account, {"name": "Server Revoke Token"})
        return super().revoke_token(token)



from typing import List, Optional

from specklepy.core.api.operations import deserialize as core_deserialize
from specklepy.core.api.operations import receive as _untracked_receive
from specklepy.core.api.operations import send as core_send
from specklepy.core.api.operations import serialize as core_serialize
from specklepy.logging import metrics
from specklepy.objects.base import Base
from specklepy.transports.abstract_transport import AbstractTransport


def send(
    base: Base,
    transports: Optional[List[AbstractTransport]] = None,
    use_default_cache: bool = True,
):
    """Sends an object via the provided transports. Defaults to the local cache.

    Arguments:
        obj {Base} -- the object you want to send
        transports {list} -- where you want to send them
        use_default_cache {bool} -- toggle for the default cache.
        If set to false, it will only send to the provided transports

    Returns:
        str -- the object id of the sent object
    """
    if transports is None:
        metrics.track(metrics.SEND)
    else:
        metrics.track(metrics.SEND, getattr(transports[0], "account", None))

    return core_send(base, transports, use_default_cache)


def receive(
    obj_id: str,
    remote_transport: Optional[AbstractTransport] = None,
    local_transport: Optional[AbstractTransport] = None,
) -> Base:
    """Receives an object from a transport.

    Arguments:
        obj_id {str} -- the id of the object to receive
        remote_transport {Transport} -- the transport to receive from
        local_transport {Transport} -- the local cache to check for existing objects
                                       (defaults to `SQLiteTransport`)

    Returns:
        Base -- the base object
    """
    metrics.track(metrics.RECEIVE, getattr(remote_transport, "account", None))
    return _untracked_receive(obj_id, remote_transport, local_transport)


def serialize(base: Base, write_transports: List[AbstractTransport] = []) -> str:
    """
    Serialize a base object. If no write transports are provided,
    the object will be serialized
    without detaching or chunking any of the attributes.

    Arguments:
        base {Base} -- the object to serialize
        write_transports {List[AbstractTransport]}
        -- optional: the transports to write to

    Returns:
        str -- the serialized object
    """
    metrics.track(metrics.SDK, custom_props={"name": "Serialize"})
    return core_serialize(base, write_transports)


def deserialize(
    obj_string: str, read_transport: Optional[AbstractTransport] = None
) -> Base:
    """
    Deserialize a string object into a Base object.

    If the object contains referenced child objects that are not stored in the local db,
    a read transport needs to be provided in order to recompose
    the base with the children objects.

    Arguments:
        obj_string {str} -- the string object to deserialize
        read_transport {AbstractTransport}
            -- the transport to fetch children objects from
                (defaults to SQLiteTransport)

    Returns:
        Base -- the deserialized object
    """
    metrics.track(metrics.SDK, custom_props={"name": "Deserialize"})
    return core_deserialize(obj_string, read_transport)


__all__ = ["receive", "send", "serialize", "deserialize"]



from datetime import datetime
from typing import List, Optional

from deprecated import deprecated
from pydantic import BaseModel, Field

FE1_DEPRECATION_REASON = "Stream/Branch/Commit API is now deprecated, Use the new Project/Model/Version API functions in Client}"
FE1_DEPRECATION_VERSION = "2.20"


@deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
class Collaborator(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    role: Optional[str] = None
    avatar: Optional[str] = None


@deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
class Commit(BaseModel):
    id: Optional[str] = None
    message: Optional[str] = None
    authorName: Optional[str] = None
    authorId: Optional[str] = None
    authorAvatar: Optional[str] = None
    branchName: Optional[str] = None
    createdAt: Optional[datetime] = None
    sourceApplication: Optional[str] = None
    referencedObject: Optional[str] = None
    totalChildrenCount: Optional[int] = None
    parents: Optional[List[str]] = None

    def __repr__(self) -> str:
        return (
            f"Commit( id: {self.id}, message: {self.message}, referencedObject:"
            f" {self.referencedObject}, authorName: {self.authorName}, branchName:"
            f" {self.branchName}, createdAt: {self.createdAt} )"
        )

    def __str__(self) -> str:
        return self.__repr__()


@deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
class Commits(BaseModel):
    totalCount: Optional[int] = None
    cursor: Optional[datetime] = None
    items: List[Commit] = []


@deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
class Object(BaseModel):
    id: Optional[str] = None
    speckleType: Optional[str] = None
    applicationId: Optional[str] = None
    totalChildrenCount: Optional[int] = None
    createdAt: Optional[datetime] = None


@deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
class Branch(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    commits: Optional[Commits] = None


@deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
class Branches(BaseModel):
    totalCount: Optional[int] = None
    cursor: Optional[datetime] = None
    items: List[Branch] = []


@deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
class Stream(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    role: Optional[str] = None
    isPublic: Optional[bool] = None
    description: Optional[str] = None
    createdAt: Optional[datetime] = None
    updatedAt: Optional[datetime] = None
    collaborators: List[Collaborator] = Field(default_factory=list)
    branches: Optional[Branches] = None
    commit: Optional[Commit] = None
    object: Optional[Object] = None
    commentCount: Optional[int] = None
    favoritedDate: Optional[datetime] = None
    favoritesCount: Optional[int] = None

    def __repr__(self):
        return (
            f"Stream( id: {self.id}, name: {self.name}, description:"
            f" {self.description}, isPublic: {self.isPublic})"
        )

    def __str__(self) -> str:
        return self.__repr__()


@deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
class Streams(BaseModel):
    totalCount: Optional[int] = None
    cursor: Optional[datetime] = None
    items: List[Stream] = []


@deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
class Activity(BaseModel):
    actionType: Optional[str] = None
    info: Optional[dict] = None
    userId: Optional[str] = None
    streamId: Optional[str] = None
    resourceId: Optional[str] = None
    resourceType: Optional[str] = None
    message: Optional[str] = None
    time: Optional[datetime] = None

    def __repr__(self) -> str:
        return (
            f"Activity( streamId: {self.streamId}, actionType: {self.actionType},"
            f" message: {self.message}, userId: {self.userId} )"
        )

    def __str__(self) -> str:
        return self.__repr__()


@deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
class ActivityCollection(BaseModel):
    totalCount: Optional[int] = None
    items: Optional[List[Activity]] = None
    cursor: Optional[datetime] = None

    def __repr__(self) -> str:
        return (
            f"ActivityCollection( totalCount: {self.totalCount}, items:"
            f" {len(self.items) if self.items else 0}, cursor:"
            f" {self.cursor.isoformat() if self.cursor else None} )"
        )

    def __str__(self) -> str:
        return self.__repr__()



from datetime import datetime, timezone
from typing import List, Optional

from deprecated import deprecated
from gql import gql

from specklepy.core.api.models import (
    ActivityCollection,
    PendingStreamCollaborator,
    Stream,
)
from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
)
from specklepy.core.api.resource import ResourceBase
from specklepy.logging.exceptions import SpeckleException, UnsupportedException

NAME = "stream"


class Resource(ResourceBase):
    """
    API Access class for streams
    Stream resource is deprecated, please use project resource instead
    """

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )

        self.schema = Stream

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get(self, id: str, branch_limit: int = 10, commit_limit: int = 10) -> Stream:
        """Get the specified stream from the server

        Arguments:
            id {str} -- the stream id
            branch_limit {int} -- the maximum number of branches to return
            commit_limit {int} -- the maximum number of commits to return

        Returns:
            Stream -- the retrieved stream
        """
        query = gql(
            """
            query Stream($id: String!, $branch_limit: Int!, $commit_limit: Int!) {
                stream(id: $id) {
                    id
                    name
                    role
                    description
                    isPublic
                    createdAt
                    updatedAt
                    commentCount
                    favoritesCount
                    collaborators {
                        id
                        name
                        role
                        avatar
                    }
                    branches(limit: $branch_limit) {
                        totalCount
                        cursor
                        items {
                            id
                            name
                            description
                            commits(limit: $commit_limit) {
                                totalCount
                                cursor
                                items {
                                    id
                                    message
                                    authorId
                                    createdAt
                                    authorName
                                    referencedObject
                                    sourceApplication
                                }
                              }
                          }
                      }
                }
            }
          """
        )

        params = {"id": id, "branch_limit": branch_limit, "commit_limit": commit_limit}

        return self.make_request(query=query, params=params, return_type="stream")

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def list(self, stream_limit: int = 10) -> List[Stream]:
        """Get a list of the user's streams

        Arguments:
            stream_limit {int} -- The maximum number of streams to return

        Returns:
            List[Stream] -- A list of Stream objects
        """
        query = gql(
            """
            query User($stream_limit: Int!) {
                user {
                    id
                    bio
                    name
                    email
                    avatar
                    company
                    verified
                    profiles
                    role
                    streams(limit: $stream_limit) {
                        totalCount
                        cursor
                        items {
                            id
                            name
                            role
                            isPublic
                            createdAt
                            updatedAt
                            description
                            commentCount
                            favoritesCount
                            collaborators {
                                id
                                name
                                role
                            }
                          }
                      }
                  }
            }
          """
        )

        params = {"stream_limit": stream_limit}

        return self.make_request(
            query=query, params=params, return_type=["user", "streams", "items"]
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def create(
        self,
        name: str = "Anonymous Python Stream",
        description: str = "No description provided",
        is_public: bool = True,
    ) -> str:
        """Create a new stream

        Arguments:
            name {str} -- the name of the string
            description {str} -- a short description of the stream
            is_public {bool}
                -- whether or not the stream can be viewed by anyone with the id

        Returns:
            id {str} -- the id of the newly created stream
        """
        query = gql(
            """
            mutation StreamCreate($stream: StreamCreateInput!) {
              streamCreate(stream: $stream)
            }
        """
        )
        if len(name) < 3 and len(name) != 0:
            return SpeckleException(message="Stream Name must be at least 3 characters")
        params = {
            "stream": {"name": name, "description": description, "isPublic": is_public}
        }

        return self.make_request(
            query=query, params=params, return_type="streamCreate", parse_response=False
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def update(
        self,
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_public: Optional[bool] = None,
    ) -> bool:
        """Update an existing stream

        Arguments:
            id {str} -- the id of the stream to be updated
            name {str} -- the name of the string
            description {str} -- a short description of the stream
            is_public {bool}
                -- whether or not the stream can be viewed by anyone with the id

        Returns:
            bool -- whether the stream update was successful
        """
        query = gql(
            """
            mutation StreamUpdate($stream: StreamUpdateInput!) {
                streamUpdate(stream: $stream)
            }
        """
        )

        params = {
            "id": id,
            "name": name,
            "description": description,
            "isPublic": is_public,
        }
        # remove None values so graphql doesn't cry
        params = {"stream": {k: v for k, v in params.items() if v is not None}}

        return self.make_request(
            query=query, params=params, return_type="streamUpdate", parse_response=False
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def delete(self, id: str) -> bool:
        """Delete a stream given its id

        Arguments:
            id {str} -- the id of the stream to delete

        Returns:
            bool -- whether the deletion was successful
        """
        query = gql(
            """
            mutation StreamDelete($id: String!) {
                streamDelete(id: $id)
            }
            """
        )

        params = {"id": id}

        return self.make_request(
            query=query, params=params, return_type="streamDelete", parse_response=False
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def search(
        self,
        search_query: str,
        limit: int = 25,
        branch_limit: int = 10,
        commit_limit: int = 10,
    ):
        """Search for streams by name, description, or id

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
            branch_limit {int} -- the maximum number of branches to return
            commit_limit {int} -- the maximum number of commits to return

        Returns:
            List[Stream] -- a list of Streams that match the search query
        """
        query = gql(
            """
            query StreamSearch(
                $search_query: String!,
                $limit: Int!,
                $branch_limit:Int!,
                $commit_limit:Int!
            ) {
                streams(query: $search_query, limit: $limit) {
                    items {
                        id
                        name
                        role
                        description
                        isPublic
                        createdAt
                        updatedAt
                        collaborators {
                            id
                            name
                            role
                            avatar
                        }
                        branches(limit: $branch_limit) {
                            totalCount
                            cursor
                            items {
                                id
                                name
                                description
                                commits(limit: $commit_limit) {
                                    totalCount
                                    cursor
                                    items {
                                        id
                                        referencedObject
                                        message
                                        authorName
                                        authorId
                                        createdAt
                                    }
                                }
                            }
                        }
                    }
                }
            }
          """
        )

        params = {
            "search_query": search_query,
            "limit": limit,
            "branch_limit": branch_limit,
            "commit_limit": commit_limit,
        }

        return self.make_request(
            query=query, params=params, return_type=["streams", "items"]
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def favorite(self, stream_id: str, favorited: bool = True):
        """Favorite or unfavorite the given stream.

        Arguments:
            stream_id {str} -- the id of the stream to favorite / unfavorite
            favorited {bool}
                -- whether to favorite (True) or unfavorite (False) the stream

        Returns:
            Stream -- the stream with its `id`, `name`, and `favoritedDate`
        """
        query = gql(
            """
            mutation StreamFavorite($stream_id: String!, $favorited: Boolean!) {
                streamFavorite(streamId: $stream_id, favorited: $favorited) {
                    id
                    name
                    favoritedDate
                    favoritesCount
                }
            }
            """
        )

        params = {
            "stream_id": stream_id,
            "favorited": favorited,
        }

        return self.make_request(
            query=query, params=params, return_type=["streamFavorite"]
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get_all_pending_invites(
        self, stream_id: str
    ) -> List[PendingStreamCollaborator]:
        """Get all of the pending invites on a stream.
        You must be a `stream:owner` to query this.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the stream id from which to get the pending invites

        Returns:
            List[PendingStreamCollaborator]
                -- a list of pending invites for the specified stream
        """
        self._check_invites_supported()

        query = gql(
            """
            query StreamInvites($streamId: String!) {
                stream(id: $streamId){
                    pendingCollaborators {
                        id
                        token
                        inviteId
                        streamId
                        streamName
                        projectName
                        projectId
                        title
                        role
                        invitedBy{
                            id
                            name
                            bio
                            company
                            avatar
                            verified
                            role
                        }
                        user {
                            id
                            name
                            bio
                            company
                            avatar
                            verified
                            role
                        }
                    }
                }
            }
            """
        )
        params = {"streamId": stream_id}

        return self.make_request(
            query=query,
            params=params,
            return_type=["stream", "pendingCollaborators"],
            schema=PendingStreamCollaborator,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def invite(
        self,
        stream_id: str,
        email: Optional[str] = None,
        user_id: Optional[str] = None,
        role: str = "stream:contributor",  # should default be reviewer?
        message: Optional[str] = None,
    ):
        """Invite someone to a stream using either their email or user id

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to invite the user to
            email {str} -- the email of the user to invite (use this OR `user_id`)
            user_id {str} -- the id of the user to invite (use this OR `email`)
            role {str}
                -- the role to assign to the user (defaults to `stream:contributor`)
            message {str}
                -- a message to send along with this invite to the specified user

        Returns:
            bool -- True if the operation was successful
        """
        self._check_invites_supported()

        if email is None and user_id is None:
            raise SpeckleException(
                "You must provide either an email or a user id to use the"
                " `stream.invite` method"
            )

        query = gql(
            """
            mutation StreamInviteCreate($input: StreamInviteCreateInput!) {
                streamInviteCreate(input: $input)
            }
            """
        )

        params = {
            "email": email,
            "userId": user_id,
            "streamId": stream_id,
            "message": message,
            "role": role,
        }
        params = {"input": {k: v for k, v in params.items() if v is not None}}

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInviteCreate",
            parse_response=False,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def invite_batch(
        self,
        stream_id: str,
        emails: Optional[List[str]] = None,
        user_ids: Optional[List[None]] = None,
        message: Optional[str] = None,
    ) -> bool:
        """Invite a batch of users to a specified stream.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to invite the user to
            emails {List[str]}
                -- the email of the user to invite (use this and/or `user_ids`)
            user_id {List[str]}
                -- the id of the user to invite (use this and/or `emails`)
            message {str}
                -- a message to send along with this invite to the specified user

        Returns:
            bool -- True if the operation was successful
        """
        self._check_invites_supported()
        if emails is None and user_ids is None:
            raise SpeckleException(
                "You must provide either an email or a user id to use the"
                " `stream.invite` method"
            )

        query = gql(
            """
            mutation StreamInviteBatchCreate($input: [StreamInviteCreateInput!]!) {
                streamInviteBatchCreate(input: $input)
            }
            """
        )

        email_invites = [
            {"streamId": stream_id, "message": message, "email": email}
            for email in (emails if emails is not None else [])
            if email is not None
        ]

        user_invites = [
            {"streamId": stream_id, "message": message, "userId": user_id}
            for user_id in (user_ids if user_ids is not None else [])
            if user_id is not None
        ]

        params = {"input": [*email_invites, *user_invites]}

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInviteBatchCreate",
            parse_response=False,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def invite_cancel(self, stream_id: str, invite_id: str) -> bool:
        """Cancel an existing stream invite

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream invite
            invite_id {str} -- the id of the invite to use

        Returns:
            bool -- true if the operation was successful
        """
        self._check_invites_supported()

        query = gql(
            """
            mutation StreamInviteCancel($streamId: String!, $inviteId: String!) {
                streamInviteCancel(streamId: $streamId, inviteId: $inviteId)
            }
            """
        )

        params = {"streamId": stream_id, "inviteId": invite_id}

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInviteCancel",
            parse_response=False,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def invite_use(self, stream_id: str, token: str, accept: bool = True) -> bool:
        """Accept or decline a stream invite

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str}
                -- the id of the stream for which the user has a pending invite
            token {str} -- the token of the invite to use
            accept {bool} -- whether or not to accept the invite (defaults to True)

        Returns:
            bool -- true if the operation was successful
        """
        self._check_invites_supported()

        query = gql(
            """
            mutation StreamInviteUse(
                $accept: Boolean!,
                $streamId: String!,
                $token: String!
                ) {
                streamInviteUse(accept: $accept, streamId: $streamId, token: $token)
            }
            """
        )

        params = {"streamId": stream_id, "token": token, "accept": accept}

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInviteUse",
            parse_response=False,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def update_permission(self, stream_id: str, user_id: str, role: str):
        """Updates permissions for a user on a given stream

        Valid for Speckle Server >=2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to grant permissions to
            user_id {str} -- the id of the user to grant permissions for
            role {str} -- the role to grant the user

        Returns:
            bool -- True if the operation was successful
        """
        if self.server_version and (
            self.server_version != ("dev",) and self.server_version < (2, 6, 4)
        ):
            raise UnsupportedException(
                "Server mutation `update_permission` is only supported as of Speckle"
                " Server v2.6.4. Please update your Speckle Server to use this method"
                " or use the `grant_permission` method instead."
            )
        query = gql(
            """
            mutation StreamUpdatePermission(
                $permission_params: StreamUpdatePermissionInput!
                ) {
                streamUpdatePermission(permissionParams: $permission_params)
            }
            """
        )

        params = {
            "permission_params": {
                "streamId": stream_id,
                "userId": user_id,
                "role": role,
            }
        }

        return self.make_request(
            query=query,
            params=params,
            return_type="streamUpdatePermission",
            parse_response=False,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def revoke_permission(self, stream_id: str, user_id: str):
        """Revoke permissions from a user on a given stream

        Arguments:
            stream_id {str} -- the id of the stream to revoke permissions from
            user_id {str} -- the id of the user to revoke permissions from

        Returns:
            bool -- True if the operation was successful
        """
        query = gql(
            """
            mutation StreamRevokePermission(
                $permission_params: StreamRevokePermissionInput!
                ) {
                streamRevokePermission(permissionParams: $permission_params)
            }
            """
        )

        params = {"permission_params": {"streamId": stream_id, "userId": user_id}}

        return self.make_request(
            query=query,
            params=params,
            return_type="streamRevokePermission",
            parse_response=False,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def activity(
        self,
        stream_id: str,
        action_type: Optional[str] = None,
        limit: int = 20,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ):
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.

        Note: all timestamps arguments should be `datetime` of any tz
            as they will be converted to UTC ISO format strings

        stream_id {str} -- the id of the stream to get activity from
        action_type {str}
            -- filter results to a single action type
            (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime}
            -- latest cutoff for activity (ie: return all activity _before_ this time)
        after {datetime}
            -- oldest cutoff for activity (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """
        query = gql(
            """
            query StreamActivity(
                $stream_id: String!,
                $action_type: String,
                $before:DateTime,
                $after: DateTime,
                $cursor: DateTime,
                $limit: Int
                ){
                stream(id: $stream_id) {
                    activity(
                        actionType: $action_type,
                        before: $before,
                        after: $after,
                        cursor: $cursor,
                        limit: $limit
                        ) {
                        totalCount
                        cursor
                        items {
                            actionType
                            info
                            userId
                            streamId
                            resourceId
                            resourceType
                            message
                            time
                        }
                    }
                }
            }
            """
        )
        try:
            params = {
                "stream_id": stream_id,
                "limit": limit,
                "action_type": action_type,
                "before": (
                    before.astimezone(timezone.utc).isoformat() if before else before
                ),
                "after": after.astimezone(timezone.utc).isoformat() if after else after,
                "cursor": (
                    cursor.astimezone(timezone.utc).isoformat() if cursor else cursor
                ),
            }
        except AttributeError as e:
            raise SpeckleException(
                "Could not get stream activity - `before`, `after`, and `cursor` must"
                " be in `datetime` format if provided",
                ValueError(),
            ) from e

        return self.make_request(
            query=query,
            params=params,
            return_type=["stream", "activity"],
            schema=ActivityCollection,
        )



from deprecated import deprecated

from specklepy.core.api.models.deprecated import FE1_DEPRECATION_VERSION
from specklepy.core.api.resources import ActiveUserResource


@deprecated(
    reason="Class renamed to ActiveUserResource", version=FE1_DEPRECATION_VERSION
)
class Resource(ActiveUserResource):
    """
    Class renamed to ActiveUserResource
    """

    pass



"""Builtin Speckle object kit."""

from specklepy.objects.GIS.CRS import CRS
from specklepy.objects.GIS.geometry import (
    GisLineElement,
    GisPointElement,
    GisPolygonElement,
    GisPolygonGeometry,
    GisRasterElement,
    PolygonGeometry,
)
from specklepy.objects.GIS.layers import RasterLayer, VectorLayer

__all__ = [
    "VectorLayer",
    "RasterLayer",
    "GisPolygonGeometry",
    "PolygonGeometry",
    "GisPolygonElement",
    "GisLineElement",
    "GisPointElement",
    "GisRasterElement",
    "CRS",
]



from typing import Dict, List

from deprecated import deprecated

from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
)
from specklepy.core.api.resources.deprecated.object import Resource as CoreResource
from specklepy.logging import metrics
from specklepy.objects.base import Base


class Resource(CoreResource):
    """API Access class for objects"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
        )
        self.schema = Base

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get(self, stream_id: str, object_id: str) -> Base:
        """
        Get a stream object

        Arguments:
            stream_id {str} -- the id of the stream for the object
            object_id {str} -- the hash of the object you want to get

        Returns:
            Base -- the returned Base object
        """
        metrics.track(metrics.SDK, self.account, {"name": "Object Get"})
        return super().get(stream_id, object_id)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def create(self, stream_id: str, objects: List[Dict]) -> str:
        """
        Not advised - generally, you want to use `operations.send()`.

        Create a new object on a stream.
        To send a base object, you can prepare it by running it through the
        `BaseObjectSerializer.traverse_base()` function to get a valid (serialisable)
        object to send.

        NOTE: this does not create a commit - you can create one with
        `SpeckleClient.commit.create`.
        Dynamic fields will be located in the 'data' dict of the received `Base` object

        Arguments:
            stream_id {str} -- the id of the stream you want to send the object to
            objects {List[Dict]}
            -- a list of base dictionary objects (NOTE: must be json serialisable)

        Returns:
            str -- the id of the object
        """
        metrics.track(metrics.SDK, self.account, {"name": "Object Create"})
        return super().create(stream_id, objects)



from typing import List, Optional, Union

from deprecated import deprecated

from specklepy.api.models import Commit
from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
)
from specklepy.core.api.resources.deprecated.commit import Resource as CoreResource
from specklepy.logging import metrics
from specklepy.logging.exceptions import SpeckleException


class Resource(CoreResource):
    """API Access class for commits"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
        )
        self.schema = Commit

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get(self, stream_id: str, commit_id: str) -> Commit:
        """
        Gets a commit given a stream and the commit id

        Arguments:
            stream_id {str} -- the stream where we can find the commit
            commit_id {str} -- the id of the commit you want to get

        Returns:
            Commit -- the retrieved commit object
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit Get"})
        return super().get(stream_id, commit_id)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def list(self, stream_id: str, limit: int = 10) -> List[Commit]:
        """
        Get a list of commits on a given stream

        Arguments:
            stream_id {str} -- the stream where the commits are
            limit {int} -- the maximum number of commits to fetch (default = 10)

        Returns:
            List[Commit] -- a list of the most recent commit objects
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit List"})
        return super().list(stream_id, limit)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def create(
        self,
        stream_id: str,
        object_id: str,
        branch_name: str = "main",
        message: str = "",
        source_application: str = "python",
        parents: Optional[List[str]] = None,
    ) -> Union[str, SpeckleException]:
        """
        Creates a commit on a branch

        Arguments:
            stream_id {str} -- the stream you want to commit to
            object_id {str} -- the hash of your commit object
            branch_name {str}
                -- the name of the branch to commit to (defaults to "main")
            message {str}
                -- optional: a message to give more information about the commit
            source_application{str}
                -- optional: the application from which the commit was created
                (defaults to "python")
            parents {List[str]} -- optional: the id of the parent commits

        Returns:
            str -- the id of the created commit
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit Create"})
        return super().create(
            stream_id, object_id, branch_name, message, source_application, parents
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def update(self, stream_id: str, commit_id: str, message: str) -> bool:
        """
        Update a commit

        Arguments:
            stream_id {str}
                -- the id of the stream that contains the commit you'd like to update
            commit_id {str} -- the id of the commit you'd like to update
            message {str} -- the updated commit message

        Returns:
            bool -- True if the operation succeeded
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit Update"})
        return super().update(stream_id, commit_id, message)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def delete(self, stream_id: str, commit_id: str) -> bool:
        """
        Delete a commit

        Arguments:
            stream_id {str}
                -- the id of the stream that contains the commit you'd like to delete
            commit_id {str} -- the id of the commit you'd like to delete

        Returns:
            bool -- True if the operation succeeded
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit Delete"})
        return super().delete(stream_id, commit_id)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def received(
        self,
        stream_id: str,
        commit_id: str,
        source_application: str = "python",
        message: Optional[str] = None,
    ) -> bool:
        """
        Mark a commit object a received by the source application.
        """
        metrics.track(metrics.SDK, self.account, {"name": "Commit Received"})
        return super().received(stream_id, commit_id, source_application, message)



from typing import Optional, Sequence

from pydantic import BaseModel


class CreateModelInput(BaseModel):
    name: str
    description: Optional[str] = None
    projectId: str


class DeleteModelInput(BaseModel):
    id: str
    projectId: str


class UpdateModelInput(BaseModel):
    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    projectId: str


class ModelVersionsFilter(BaseModel):
    priorityIds: Sequence[str]
    priorityIdsOnly: Optional[bool] = None



from typing import Optional

from specklepy.core.api.inputs.project_inputs import (
    ProjectCreateInput,
    ProjectModelsFilter,
    ProjectUpdateInput,
    ProjectUpdateRoleInput,
)
from specklepy.core.api.models import Project, ProjectWithModels, ProjectWithTeam
from specklepy.core.api.resources import ProjectResource as CoreResource
from specklepy.logging import metrics


class ProjectResource(CoreResource):
    """API Access class for projects"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )

    def get(self, project_id: str) -> Project:
        metrics.track(metrics.SDK, self.account, {"name": "Project Get "})
        return super().get(project_id)

    def get_with_models(
        self,
        project_id: str,
        *,
        models_limit: int = 25,
        models_cursor: Optional[str] = None,
        models_filter: Optional[ProjectModelsFilter] = None,
    ) -> ProjectWithModels:
        metrics.track(metrics.SDK, self.account, {"name": "Project Get With Models"})
        return super().get_with_models(
            project_id,
            models_limit=models_limit,
            models_cursor=models_cursor,
            models_filter=models_filter,
        )

    def get_with_team(self, project_id: str) -> ProjectWithTeam:
        metrics.track(metrics.SDK, self.account, {"name": "Project Get With Team"})
        return super().get_with_team(project_id)

    def create(self, input: ProjectCreateInput) -> Project:
        metrics.track(metrics.SDK, self.account, {"name": "Project Create"})
        return super().create(input)

    def update(self, input: ProjectUpdateInput) -> Project:
        metrics.track(metrics.SDK, self.account, {"name": "Project Update"})
        return super().update(input)

    def delete(self, project_id: str) -> bool:
        metrics.track(metrics.SDK, self.account, {"name": "Project Delete"})
        return super().delete(project_id)

    def update_role(self, input: ProjectUpdateRoleInput) -> ProjectWithTeam:
        metrics.track(metrics.SDK, self.account, {"name": "Project Update Role"})
        return super().update_role(input)



import re
from typing import Any, Dict, List, Tuple

import requests
from gql import gql

from specklepy.core.api.models import ServerInfo
from specklepy.core.api.resource import ResourceBase
from specklepy.logging.exceptions import GraphQLException

NAME = "server"


class ServerResource(ResourceBase):
    """API Access class for the server"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
        )

    def get(self) -> ServerInfo:
        """Get the server info

        Returns:
            dict -- the server info in dictionary form
        """
        query = gql(
            """
            query Server {
                serverInfo {
                    name
                    company
                    description
                    adminContact
                    canonicalUrl
                    version
                    roles {
                        name
                        description
                        resourceTarget
                    }
                    scopes {
                        name
                        description
                    }
                    authStrategies{
                        id
                        name
                        icon
                    }
                }
            }
            """
        )

        server_info = self.make_request(
            query=query, return_type="serverInfo", schema=ServerInfo
        )
        if isinstance(server_info, ServerInfo) and isinstance(
            server_info.canonicalUrl, str
        ):
            r = requests.get(
                server_info.canonicalUrl, headers={"User-Agent": "specklepy SDK"}
            )
            if "x-speckle-frontend-2" in r.headers:
                server_info.frontend2 = True
            else:
                server_info.frontend2 = False

        return server_info

    def version(self) -> Tuple[Any, ...]:
        """Get the server version

        Returns:
            the server version in the format (major, minor, patch, (tag, build))
            eg (2, 6, 3) for a stable build and (2, 6, 4, 'alpha', 4711) for alpha
        """
        # not tracking as it will be called along with other mutations / queries as a check
        query = gql(
            """
            query Server {
                serverInfo {
                    version
                }
            }
            """
        )
        ver = self.make_request(
            query=query, return_type=["serverInfo", "version"], parse_response=False
        )
        if isinstance(ver, Exception):
            raise GraphQLException(
                f"Could not get server version for {self.basepath}", [ver]
            )

        # pylint: disable=consider-using-generator; (list comp is faster)
        return tuple(
            [
                int(segment) if segment.isdigit() else segment
                for segment in re.split(r"\.|-", ver)
            ]
        )

    def apps(self) -> Dict:
        """Get the apps registered on the server

        Returns:
            dict -- a dictionary of apps registered on the server
        """
        query = gql(
            """
            query Apps {
                apps{
                    id
                    name
                    description
                    termsAndConditionsLink
                    trustByDefault
                    logo
                    author {
                        id
                        name
                        avatar
                    }
                }
            }
        """
        )

        return self.make_request(query=query, return_type="apps", parse_response=False)

    def create_token(self, name: str, scopes: List[str], lifespan: int) -> str:
        """Create a personal API token

        Arguments:
            scopes {List[str]} -- the scopes to grant with this token
            name {str} -- a name for your new token
            lifespan {int} -- duration before the token expires

        Returns:
            str -- the new API token. note: this is the only time you'll see the token!
        """
        query = gql(
            """
            mutation TokenCreate($token: ApiTokenCreateInput!) {
                apiTokenCreate(token: $token)
            }
            """
        )
        params = {"token": {"scopes": scopes, "name": name, "lifespan": lifespan}}

        return self.make_request(
            query=query,
            params=params,
            return_type="apiTokenCreate",
            parse_response=False,
        )

    def revoke_token(self, token: str) -> bool:
        """Revokes (deletes) a personal API token

        Arguments:
            token {str} -- the token to revoke (delete)

        Returns:
            bool -- True if the token was successfully deleted
        """
        query = gql(
            """
            mutation TokenRevoke($token: String!) {
                apiTokenRevoke(token: $token)
            }
            """
        )
        params = {"token": token}

        return self.make_request(
            query=query,
            params=params,
            return_type="apiTokenRevoke",
            parse_response=False,
        )



from specklepy.core.api.models.current import (
    AuthStrategy,
    LimitedUser,
    Model,
    ModelWithVersions,
    PendingStreamCollaborator,
    Project,
    ProjectCollaborator,
    ProjectCommentCollection,
    ProjectWithModels,
    ProjectWithTeam,
    ResourceCollection,
    ServerConfiguration,
    ServerInfo,
    ServerMigration,
    User,
    UserSearchResultCollection,
    Version,
)
from specklepy.core.api.models.deprecated import (
    Activity,
    ActivityCollection,
    Branch,
    Branches,
    Collaborator,
    Commit,
    Commits,
    Object,
    Stream,
    Streams,
)
from specklepy.core.api.models.subscription_messages import (
    ProjectModelsUpdatedMessage,
    ProjectUpdatedMessage,
    ProjectVersionsUpdatedMessage,
    UserProjectsUpdatedMessage,
)

__all__ = [
    "User",
    "ResourceCollection",
    "ServerMigration",
    "AuthStrategy",
    "ServerConfiguration",
    "ServerInfo",
    "LimitedUser",
    "PendingStreamCollaborator",
    "ProjectCollaborator",
    "Version",
    "Model",
    "ModelWithVersions",
    "Project",
    "ProjectWithModels",
    "ProjectWithTeam",
    "ProjectCommentCollection",
    "UserSearchResultCollection",
    "UserProjectsUpdatedMessage",
    "ProjectModelsUpdatedMessage",
    "ProjectUpdatedMessage",
    "ProjectVersionsUpdatedMessage",
    "Collaborator",
    "Commit",
    "Commits",
    "Object",
    "Branch",
    "Branches",
    "Stream",
    "Streams",
    "Activity",
    "ActivityCollection",
]



from typing import Callable, Optional, Sequence

from pydantic import BaseModel
from typing_extensions import TypeVar

from specklepy.core.api.models import (
    ProjectModelsUpdatedMessage,
    ProjectUpdatedMessage,
    ProjectVersionsUpdatedMessage,
    UserProjectsUpdatedMessage,
)
from specklepy.core.api.resources import SubscriptionResource as CoreResource
from specklepy.logging import metrics

TEventArgs = TypeVar("TEventArgs", bound=BaseModel)


class SubscriptionResource(CoreResource):
    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
        )

    async def user_projects_updated(
        self, callback: Callable[[UserProjectsUpdatedMessage], None]
    ) -> None:
        metrics.track(
            metrics.SDK, self.account, {"name": "Subscription Project Models Updated"}
        )
        return await super().user_projects_updated(callback)

    async def project_models_updated(
        self,
        callback: Callable[[ProjectModelsUpdatedMessage], None],
        id: str,
        *,
        model_ids: Optional[Sequence[str]] = None,
    ) -> None:
        metrics.track(
            metrics.SDK, self.account, {"name": "Subscription Project Models Updated"}
        )
        return await super().project_models_updated(callback, id, model_ids=model_ids)

    async def project_updated(
        self,
        callback: Callable[[ProjectUpdatedMessage], None],
        id: str,
    ) -> None:
        metrics.track(
            metrics.SDK, self.account, {"name": "Subscription Project Updated"}
        )
        return await super().project_updated(callback, id)

    async def project_versions_updated(
        self,
        callback: Callable[[ProjectVersionsUpdatedMessage], None],
        id: str,
    ) -> None:
        metrics.track(
            metrics.SDK, self.account, {"name": "Subscription Project Versions Updated"}
        )
        return await super().project_versions_updated(callback, id)



from typing import Callable, Dict, List, Optional, Union

from deprecated import deprecated
from graphql import DocumentNode

from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
)
from specklepy.core.api.resources.current.subscription_resource import check_wsclient
from specklepy.core.api.resources.deprecated.subscriptions import (
    Resource as CoreResource,
)
from specklepy.logging import metrics


class Resource(CoreResource):
    """API Access class for subscriptions"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    @check_wsclient
    async def stream_added(self, callback: Optional[Callable] = None):
        """Subscribes to new stream added event for your profile.
        Use this to display an up-to-date list of streams.

        Arguments:
            callback {Callable[Stream]} -- a function that takes the updated stream
            as an argument and executes each time a stream is added

        Returns:
            Stream -- the update stream
        """
        metrics.track(metrics.SDK, self.account, {"name": "Subscription Stream Added"})
        return super().stream_added(callback)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    @check_wsclient
    async def stream_updated(self, id: str, callback: Optional[Callable] = None):
        """
        Subscribes to stream updated event.
        Use this in clients/components that pertain only to this stream.

        Arguments:
            id {str} -- the stream id of the stream to subscribe to
            callback {Callable[Stream]}
            -- a function that takes the updated stream
            as an argument and executes each time the stream is updated

        Returns:
            Stream -- the update stream
        """
        metrics.track(
            metrics.SDK, self.account, {"name": "Subscription Stream Updated"}
        )
        return super().stream_updated(id, callback)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    @check_wsclient
    async def stream_removed(self, callback: Optional[Callable] = None):
        """Subscribes to stream removed event for your profile.
        Use this to display an up-to-date list of streams for your profile.
        NOTE: If someone revokes your permissions on a stream,
        this subscription will be triggered with an extra value of revokedBy
        in the payload.

        Arguments:
            callback {Callable[Dict]}
            -- a function that takes the returned dict as an argument
            and executes each time a stream is removed

        Returns:
            dict -- dict containing 'id' of stream removed and optionally 'revokedBy'
        """
        metrics.track(
            metrics.SDK, self.account, {"name": "Subscription Stream Removed"}
        )
        return super().stream_removed(callback)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    @check_wsclient
    async def subscribe(
        self,
        query: DocumentNode,
        params: Optional[Dict] = None,
        callback: Optional[Callable] = None,
        return_type: Optional[Union[str, List]] = None,
        schema=None,
        parse_response: bool = True,
    ):
        # if self.client.transport.websocket is None:
        # TODO: add multiple subs to the same ws connection
        async with self.client as session:
            async for res in session.subscribe(query, variable_values=params):
                res = self._step_into_response(response=res, return_type=return_type)
                if parse_response:
                    res = self._parse_response(response=res, schema=schema)
                if callback is not None:
                    callback(res)
                else:
                    return res



from enum import Enum
from typing import Union

from specklepy.logging.exceptions import SpeckleException, SpeckleInvalidUnitException

__all__ = [
    "Units",
    "get_encoding_from_units",
    "get_units_from_encoding",
    "get_units_from_string",
]


class Units(Enum):
    mm = "mm"
    cm = "cm"
    m = "m"
    km = "km"
    inches = "in"
    feet = "ft"
    yards = "yd"
    miles = "mi"
    none = "none"


UNITS_STRINGS = {
    Units.mm: ["mm", "mil", "millimeters", "millimetres"],
    Units.cm: ["cm", "centimetre", "centimeter", "centimetres", "centimeters"],
    Units.m: ["m", "meter", "meters", "metre", "metres"],
    Units.km: ["km", "kilometer", "kilometre", "kilometers", "kilometres"],
    Units.inches: ["in", "inch", "inches"],
    Units.feet: ["ft", "foot", "feet"],
    Units.yards: ["yd", "yard", "yards"],
    Units.miles: ["mi", "mile", "miles"],
    Units.none: ["none", "null"],
}


UNITS_ENCODINGS = {
    Units.none: 0,
    None: 0,
    Units.mm: 1,
    Units.cm: 2,
    Units.m: 3,
    Units.km: 4,
    Units.inches: 5,
    Units.feet: 6,
    Units.yards: 7,
    Units.miles: 8,
}


UNIT_SCALE = {
    Units.none: 1,
    Units.mm: 0.001,
    Units.cm: 0.01,
    Units.m: 1.0,
    Units.km: 1000.0,
    Units.inches: 0.0254,
    Units.feet: 0.3048,
    Units.yards: 0.9144,
    Units.miles: 1609.340,
}
"""Unit scaling factor to meters"""


def get_units_from_string(unit: str) -> Units:
    if not isinstance(unit, str):
        raise SpeckleInvalidUnitException(unit)
    unit = str.lower(unit)
    for name, alternates in UNITS_STRINGS.items():
        if unit in alternates:
            return name
    raise SpeckleInvalidUnitException(unit)


def get_units_from_encoding(unit: int) -> Units:
    for name, encoding in UNITS_ENCODINGS.items():
        if unit == encoding:
            return name or Units.none

    raise SpeckleException(
        message=(
            f"Could not understand what unit {unit} is referring to."
            f"Please enter a valid unit encoding (eg {UNITS_ENCODINGS})."
        )
    )


def get_encoding_from_units(unit: Union[Units, str, None]):
    maybe_sanitized_unit = unit
    if isinstance(unit, str):
        for unit_enum, aliases in UNITS_STRINGS.items():
            if unit in aliases:
                maybe_sanitized_unit = unit_enum
    try:
        return UNITS_ENCODINGS[maybe_sanitized_unit]
    except KeyError as e:
        raise SpeckleException(
            message=(
                f"No encoding exists for unit {maybe_sanitized_unit}."
                f"Please enter a valid unit to encode (eg {UNITS_ENCODINGS})."
            )
        ) from e


def get_scale_factor_from_string(fromUnits: str, toUnits: str) -> float:
    """Returns a scalar to convert distance values from one unit system to another"""
    return get_scale_factor(
        get_units_from_string(fromUnits), get_units_from_string(toUnits)
    )


def get_scale_factor(fromUnits: Units, toUnits: Units) -> float:
    """Returns a scalar to convert distance values from one unit system to another"""
    return get_scale_factor_to_meters(fromUnits) / get_scale_factor_to_meters(toUnits)


def get_scale_factor_to_meters(fromUnits: Units) -> float:
    """Returns a scalar to convert distance values from one unit system to meters"""
    if fromUnits not in UNIT_SCALE:
        raise ValueError(f"Invalid units provided: {fromUnits}")

    return UNIT_SCALE[fromUnits]



import contextlib
import getpass
import hashlib
import logging
import platform
import queue
import sys
import threading
from typing import Optional

import requests

"""
Anonymous telemetry to help us understand how to make a better Speckle.
This really helps us to deliver a better open source project and product!
"""
TRACK = True
HOST_APP = "python"
HOST_APP_VERSION = f"python {'.'.join(map(str, sys.version_info[:2]))}"
PLATFORMS = {"win32": "Windows", "cygwin": "Windows", "darwin": "Mac OS X"}

LOG = logging.getLogger(__name__)
METRICS_TRACKER = None

# actions
SDK = "SDK Action"
CONNECTOR = "Connector Action"
RECEIVE = "Receive"
SEND = "Send"

# not in use since 2.15
ACCOUNTS = "Get Local Accounts"
BRANCH = "Branch Action"
CLIENT = "Speckle Client"
COMMIT = "Commit Action"
DESERIALIZE = "serialization/deserialize"
INVITE = "Invite Action"
OTHER_USER = "Other User Action"
PERMISSION = "Permission Action"
SERIALIZE = "serialization/serialize"
SERVER = "Server Action"
STREAM = "Stream Action"
STREAM_WRAPPER = "Stream Wrapper"
USER = "User Action"


def disable():
    global TRACK
    TRACK = False


def enable():
    global TRACK
    TRACK = True


def set_host_app(host_app: str, host_app_version: Optional[str] = None):
    global HOST_APP, HOST_APP_VERSION
    HOST_APP = host_app
    HOST_APP_VERSION = host_app_version or HOST_APP_VERSION


def track(
    action: str,
    account=None,
    custom_props: Optional[dict] = None,
):
    if not TRACK:
        return
    try:
        initialise_tracker(account)
        event_params = {
            "event": action,
            "properties": {
                "distinct_id": METRICS_TRACKER.last_user,
                "server_id": METRICS_TRACKER.last_server,
                "token": METRICS_TRACKER.analytics_token,
                "hostApp": HOST_APP,
                "hostAppVersion": HOST_APP_VERSION,
                "$os": METRICS_TRACKER.platform,
                "type": "action",
            },
        }
        if custom_props:
            event_params["properties"].update(custom_props)

        METRICS_TRACKER.queue.put_nowait(event_params)
    except Exception as ex:
        # wrapping this whole thing in a try except as we never want a failure here to annoy users!
        LOG.debug(f"Error queueing metrics request: {str(ex)}")


def initialise_tracker(account=None):
    global METRICS_TRACKER
    if not METRICS_TRACKER:
        METRICS_TRACKER = MetricsTracker()

    if account and account.userInfo.email:
        METRICS_TRACKER.set_last_user(account.userInfo.email)
    if account and account.serverInfo.url:
        METRICS_TRACKER.set_last_server(account.serverInfo.url)


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class MetricsTracker(metaclass=Singleton):
    analytics_url = "https://analytics.speckle.systems/track?ip=1"
    analytics_token = "acd87c5a50b56df91a795e999812a3a4"
    last_user = ""
    last_server = None
    platform = None
    sending_thread = None
    queue = queue.Queue(1000)

    def __init__(self) -> None:
        self.sending_thread = threading.Thread(
            target=self._send_tracking_requests, daemon=True
        )
        self.platform = PLATFORMS.get(sys.platform, "linux")
        self.sending_thread.start()
        with contextlib.suppress(Exception):
            node, user = platform.node(), getpass.getuser()
            if node and user:
                self.last_user = f"@{self.hash(f'{node}-{user}')}"

    def set_last_user(self, email: str):
        if not email:
            return
        self.last_user = f"@{self.hash(email)}"

    def set_last_server(self, server: str):
        if not server:
            return
        self.last_server = self.hash(server)

    def hash(self, value: str):
        inputList = value.lower().split("://")
        input = inputList[len(inputList) - 1].split("/")[0].split("?")[0]
        return hashlib.md5(input.encode("utf-8")).hexdigest().upper()

    def _send_tracking_requests(self):
        session = requests.Session()
        while True:
            event_params = [self.queue.get()]

            try:
                session.post(self.analytics_url, json=event_params)
            except Exception as ex:
                LOG.debug(f"Error sending metrics request: {str(ex)}")

            self.queue.task_done()



from typing import Optional

from pydantic import BaseModel


class UserUpdateInput(BaseModel):
    avatar: Optional[str] = None
    bio: Optional[str] = None
    company: Optional[str] = None
    name: Optional[str] = None
